@@ -1,7 +1,7 @@
 #! /usr/bin/env python
 # -*- coding: utf-8 -*-
 
-from __future__ import division
+from __future__ import absolute_import, division, print_function
 import urwid
 import bdb
 import sys
@@ -117,7 +117,7 @@ License:
 
 PuDB is licensed to you under the MIT/X Consortium license:
 
-Copyright (c) 2009-13 Andreas Kloeckner and contributors
+Copyright (c) 2009-16 Andreas Kloeckner and contributors
 
 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
@@ -546,6 +546,9 @@ class FileSourceCodeProvider(SourceCodeProvider):
             return [SourceLine(debugger_ui, self.file_name)]
 
         breakpoints = debugger_ui.debugger.get_file_breaks(self.file_name)[:]
+        breakpoints = [lineno for lineno in breakpoints if
+            any(bp.enabled
+                for bp in debugger_ui.debugger.get_breaks(self.file_name, lineno))]
         breakpoints += [i for f, i in debugger_ui.debugger.set_traces if f
             == self.file_name and debugger_ui.debugger.set_traces[f, i]]
         try:
@@ -984,6 +987,9 @@ class DebuggerUI(FrameVarInfoKeeper):
             bp = self._get_bp_list()[pos]
             bp.enabled = not bp.enabled
 
+            sline = self.source[bp.line-1]
+            sline.set_breakpoint(bp.enabled)
+
             self.update_breakpoints()
 
         def examine_breakpoint(w, size, key):
@@ -1205,8 +1211,17 @@ class DebuggerUI(FrameVarInfoKeeper):
                 existing_breaks = self.debugger.get_breaks(
                         bp_source_identifier, lineno)
                 if existing_breaks:
-                    err = self.debugger.clear_break(bp_source_identifier, lineno)
-                    sline.set_breakpoint(False)
+                    err = None
+                    for bp in existing_breaks:
+                        if not bp.enabled:
+                            bp.enable()
+                            sline.set_breakpoint(True)
+                            # Unsure about this. Are multiple breakpoints even
+                            # possible?
+                            break
+                    else:
+                        err = self.debugger.clear_break(bp_source_identifier, lineno)
+                        sline.set_breakpoint(False)
                 else:
                     file_lineno = (bp_source_identifier, lineno)
                     if file_lineno in self.debugger.set_traces:
@@ -1359,7 +1374,12 @@ class DebuggerUI(FrameVarInfoKeeper):
                     if widget is not new_mod_entry:
                         mod_name = widget.base_widget.get_text()[0]
                         mod = sys.modules[mod_name]
-                        reload(mod)
+                        if PY3:
+                            import importlib
+                            importlib.reload(mod)
+                        else:
+                            reload(mod)  # noqa (undef on Py3)
+
                         self.message("'%s' was successfully reloaded." % mod_name)
 
                         if self.source_code_provider is not None:
@@ -1754,11 +1774,11 @@ class DebuggerUI(FrameVarInfoKeeper):
             curframe = self.debugger.curframe
 
             import pudb.shell as shell
-            if shell.HAVE_IPYTHON and CONFIG["shell"] == "ipython":
+            if CONFIG["shell"] == "ipython" and shell.have_ipython():
                 runner = shell.run_ipython_shell
-            elif shell.HAVE_BPYTHON and CONFIG["shell"] == "bpython":
+            elif CONFIG["shell"] == "bpython" and shell.HAVE_BPYTHON:
                 runner = shell.run_bpython_shell
-            elif shell.HAVE_PTPYTHON and CONFIG["shell"] == "ptpython":
+            elif CONFIG["shell"] == "ptpython" and shell.HAVE_PTPYTHON:
                 runner = shell.run_ptpython_shell
             else:
                 runner = shell.run_classic_shell
@@ -2075,7 +2095,7 @@ class DebuggerUI(FrameVarInfoKeeper):
                 self.message("Package 'pygments' not found. "
                         "Syntax highlighting disabled.")
 
-        WELCOME_LEVEL = "e026"  # noqa
+        WELCOME_LEVEL = "e028"  # noqa
         if CONFIG["seen_welcome"] < WELCOME_LEVEL:
             CONFIG["seen_welcome"] = WELCOME_LEVEL
             from pudb import VERSION
@@ -2092,6 +2112,13 @@ class DebuggerUI(FrameVarInfoKeeper):
                     "(invoked by hitting '?' after this message) should get you "
                     "on your way.\n"
 
+                    "\nChanges in version 2017.1:\n\n"
+                    "- Many, many bug fixes (thank you to all who contributed!)\n"
+
+                    "\nChanges in version 2016.2:\n\n"
+                    "- UI improvements for disabled breakpoints.\n"
+                    "- Bug fixes.\n"
+
                     "\nChanges in version 2016.1:\n\n"
                     "- Fix module browser on Py3.\n"
 
@@ -2333,7 +2360,7 @@ class DebuggerUI(FrameVarInfoKeeper):
     def update_breakpoints(self):
         self.bp_walker[:] = [
                 BreakpointFrame(self.debugger.current_bp == (bp.file, bp.line),
-                    self._format_fname(bp.file), bp.line)
+                    self._format_fname(bp.file), bp)
                 for bp in self._get_bp_list()]
 
     def update_stack(self):
